import { colorManager } from "@/components/setting/color";
import { increaseBrightness, lowerBrightness, parseColor } from "@/hook/color";
import { globalLookup } from "@/hook/global";
import { pinkLog } from "@/hook/utils";
import axios from 'axios';
import pako from 'pako';

const mode = window.acquireVsCodeApi === undefined ? 'debug' : 'release';
pinkLog('digital-netlist-render mode: ' + mode);

let vscode = window.acquireVsCodeApi === undefined ? undefined : acquireVsCodeApi();

function getColorFromMacro(rootStyles, optionName, theme) {
    if (theme === 'dark') {
        return rootStyles.getPropertyValue(optionName);
    } else {
        switch (optionName) {
        case '--foreground':
        case '--wire-color':
        case '--cross-dot-color':
            return '#2D323B';
        }
        const color = rootStyles.getPropertyValue(optionName);
        if (!color) {
            return color;
        }
        const rgb = parseColor(color);
        const { r, g, b } = increaseBrightness(rgb, 10);
        return `rgb(${r}, ${g}, ${b})`;
    }
}

function makeStyleString(config = {}) {
    const theme = config.theme || 'dark';
    // 加入全局颜色宏
    const colorMacros = [];
    const rootStyles = getComputedStyle(document.documentElement);

    for (const item of colorManager.generals) {
        const colorOption = `--${item.type}-color`;
        const colorFillOption = `--${item.type}-fill-color`;
        
        const color = getColorFromMacro(rootStyles, colorOption, theme);
        const colorFill = getColorFromMacro(rootStyles, colorFillOption, theme);
        if (color) {
            colorMacros.push(`${colorOption}: ${color};`);
        }
        if (colorFill) {
            colorMacros.push(`${colorFillOption}: ${colorFill};`);
        }
    }

    for (const item of colorManager.cells) {
        const colorOption = `--${item.type}-color`;
        const colorFillOption = `--${item.type}-fill-color`;
        
        const color = getColorFromMacro(rootStyles, colorOption, theme);
        const colorFill = getColorFromMacro(rootStyles, colorFillOption, theme);
        if (color) {
            colorMacros.push(`${colorOption}: ${color};`);
        }
        if (colorFill) {
            colorMacros.push(`${colorFillOption}: ${colorFill};`);
        }
    }

    const globalSetting = [
        '--foreground',
        '--background',
        '--main-color'
    ];

    for (const setting of globalSetting) {
        const color = getColorFromMacro(rootStyles, setting, theme);
        if (color) {
            colorMacros.push(`${setting}: ${color};`);
        }
    }

    const styleString = `:root {\n${colorMacros.join('\n')}\n}`;
    return styleString;
}

function getCompressedSvgBuffer(config = {}) {
    const selection = globalLookup.netlistRender.selection;
    const node = selection.node();

    const styleString = makeStyleString(config);
    const style = document.createElement('style');
    style.textContent = styleString;
    node.insertBefore(style, node.firstChild);

    const serializer = new XMLSerializer();
    const svgString = serializer.serializeToString(node);
    const array = pako.gzip(svgString);

    // 还原
    const styleElement = node.querySelector('style');
    if (styleElement) {
        styleElement.remove();
    }

    return array;
}

export async function saveAsSvg() {
    const svgBuffer = getCompressedSvgBuffer();
    const moduleName = globalLookup.topModuleName;
    if (mode === 'debug') {
        const res = await axios.post('http://localhost:3000/netlist/save-as-svg', { svgBuffer, moduleName });
        if (res.success) {
            pinkLog('成功保存到”' + res.savePath);
        }
    } else {
        vscode.postMessage({
            command: 'save-as-svg',
            data: { svgBuffer, moduleName }
        });
    }
}

export async function saveAsPdf() {
    const svgBuffer = getCompressedSvgBuffer({ theme: 'light' });
    const moduleName = globalLookup.topModuleName;

    const node = globalLookup.netlistRender.selection.node();
    const width = node.getAttribute('width');
    const height = node.getAttribute('height');

    if (mode === 'debug') {
        const res = await axios.post('http://localhost:3000/netlist/save-as-pdf', { svgBuffer, moduleName, width, height });
        if (res.success) {
            pinkLog('成功保存到”' + res.savePath);
        }
    } else {
        vscode.postMessage({
            command: 'save-as-pdf',
            data: { svgBuffer, moduleName, width, height }
        });
    }
}